#include "envy.h"
#include "envy_config.h"
#include "envy_options.h"
#include "envy_subtitles.h"
#include "envy_engine.h"
#include "envy_osdb.h"

static void cb_window_resized(Ecore_Evas *ee);
static int  use_video_size(Evas_Object *o);
static int  bg_key_down(void *data, int ev_type, void *event_info);
static void cb_toggle_play_pause(void *data, Evas_Object *o, const char *emission, const char *source);
static void cb_play(void *data, Evas_Object *o, const char *emission, const char *source);
static void cb_pause(void *data, Evas_Object *o, const char *emission, const char *source);
static void cb_progress_dragged(void *data, Evas_Object *o, const char *emission, const char *source);
static void cb_frame_decoded(void *data, Evas_Object *obj, void *event_info);
static int  cb_hide_info(void *data);
static int  cb_hide_mouse(void *data);
static void cb_mouse_down(void *data, Evas *e, Evas_Object *o, void *event_info);
static void cb_mouse_move(void *data, Evas *e, Evas_Object *o, void *event_info);
static void envy_info_set(const char *msg, double time);
static void envy_info_hide(void);

int
main(int argc, char **argv)
{
  if (argc < 2)
    {
      fprintf(stderr, "No filename in arguments\n");
      printf("Usage %s filename\n", argv[0]);
      return EXIT_FAILURE;
    }

  /* Envy */
  envy = calloc(1, sizeof(Envy));
  envy->subs_size = 300;
  envy->sub_delay = 0;
  envy->show_sub = 1;

  /* Config */
  envy->config = envy_config_get();

  envy->filename = envy_parse_opt(argc, argv, envy->config);

  /* Initialization */
  if (!ecore_init()) return -1;
  edje_init();
  ecore_app_args_set(argc, (const char **)argv);

  if (!ecore_evas_init()) return -1;
  if (ecore_evas_engine_type_supported_get(envy->config->ecore_evas_engine_type))
    {
      envy->ee = envy_engine_new(envy->config->ecore_evas_engine_type,
                                 NULL, 0, 0, 0, 640, 480);
    }
  else
    {
      fprintf(stderr, "Evas %s engine isn't supported.\n",
              envy_engine_desc(envy->config->ecore_evas_engine_type));
      return -1;
    }

  if (!envy->ee)
    {
      fprintf(stderr, "Evas %s engine isn't working.\n",
              envy_engine_desc(envy->config->ecore_evas_engine_type));
      return -1;
    }

  /* Window */
  ecore_evas_title_set(envy->ee, "Envy");
  ecore_evas_name_class_set(envy->ee, "envy", "envy");
  ecore_evas_show(envy->ee);
  ecore_evas_callback_resize_set(envy->ee, cb_window_resized);

  /* Evas */
  envy->evas = ecore_evas_get(envy->ee);
  evas_image_cache_set(envy->evas, 8 * 2048);
  edje_frametime_set(1.0 / 30.0);

  /* Edje */
  envy->gui = edje_object_add(envy->evas);
  if (!access(envy->config->theme, F_OK))
    edje_object_file_set(envy->gui, envy->config->theme, "player");
  else
    fprintf(stderr, "Can't find theme file\n");
  evas_object_move(envy->gui, 0, 0);
  evas_object_resize(envy->gui, 640, 480);
  evas_object_show(envy->gui);

  /* Black Background */
  envy->black_bg = evas_object_rectangle_add(envy->evas);
  evas_object_color_set(envy->black_bg, 0, 0, 0, 255);
  evas_object_show(envy->black_bg);

  /* Balloon (logo) */
  envy->balloon = evas_object_image_add(envy->evas);
  evas_object_image_file_set(envy->balloon, PACKAGE_DATA_DIR "/pixmaps/envy.png", NULL);
  evas_object_image_fill_set(envy->balloon, 0, 0, 186, 137);
  evas_object_move(envy->balloon, 237, 182);
  evas_object_resize(envy->balloon, 186, 137);
  evas_object_show(envy->balloon);

  /* Video */
  envy->video = emotion_object_add(envy->evas);
  if (!emotion_object_init(envy->video, "xine")) return EXIT_FAILURE;
  emotion_object_file_set(envy->video, envy->filename);
  emotion_object_smooth_scale_set(envy->video, 1);
  evas_object_move(envy->video, 0, 0);
  evas_object_resize(envy->video, 800, 600);
  evas_object_show(envy->video);

  /* Resize the window at the video size */
  ecore_timer_add(0.5, (int (*)(void *))use_video_size, envy->video);

  /* Callbacks */
  edje_object_signal_callback_add(envy->gui, "play_pause_toggled", "*", cb_toggle_play_pause, envy->video);
  edje_object_signal_callback_add(envy->gui, "play", "gui", cb_play, envy->video);
  edje_object_signal_callback_add(envy->gui, "pause", "gui", cb_pause, envy->video);
  edje_object_signal_callback_add(envy->gui, "drag", "progress", cb_progress_dragged, envy->video);
  ecore_event_handler_add(ECORE_X_EVENT_KEY_DOWN, bg_key_down, envy->video);
  evas_object_smart_callback_add(envy->video, "frame_decode", cb_frame_decoded, envy->video);
  evas_object_event_callback_add(envy->video, EVAS_CALLBACK_MOUSE_MOVE, cb_mouse_move, envy->ee);
  evas_object_event_callback_add(envy->gui, EVAS_CALLBACK_MOUSE_MOVE, cb_mouse_down, envy->ee);
  evas_object_event_callback_add(envy->black_bg, EVAS_CALLBACK_MOUSE_MOVE, cb_mouse_down, envy->ee);
  evas_object_event_callback_add(envy->gui, EVAS_CALLBACK_MOUSE_DOWN, cb_mouse_down, envy->ee);

  /* Sub */
  envy->sub = evas_object_textblock_add(envy->evas);
  evas_object_color_set(envy->sub, 255, 255, 255, 255);
  evas_object_move(envy->sub, 100, 400);
  evas_object_resize(envy->sub, 800, 200);
  evas_object_pass_events_set(envy->sub, 1);
  evas_object_show(envy->sub);

  /* Textblock Style */
  Evas_Textblock_Style *style;
  char st[512];
  snprintf(st, sizeof(st),
           "DEFAULT='font=vera font_size=20 align=center valign=bottom "
           "color=#ffffff wrap=word style=glow glow_color=#000000'"
           "br='\n'");
  style = evas_textblock_style_new();
  evas_textblock_style_set(style, st);
  evas_object_textblock_style_set(envy->sub, style);

  evas_object_layer_set(envy->gui, 4);
  evas_object_layer_set(envy->sub, 3);
  evas_object_layer_set(envy->video, 2);
  evas_object_layer_set(envy->balloon, 6);
  evas_object_layer_set(envy->black_bg, 0);

  emotion_object_play_set(envy->video, 1);
  edje_object_signal_emit(envy->gui, "play", "envy");
  edje_object_signal_emit(envy->gui, "hide_info", "envy");
  ecore_main_loop_begin();
  return EXIT_SUCCESS;
}

void
envy_osdb_process(void *data)
{
  Envy_Osdb *eo = data;

  envy_info_set("Login in OSDB...", 0); 
  envy_osdb_login(eo, "en");
  envy_info_set("Searching...", 0); 
  if (envy_osdb_search(eo, eo->lang, eo->moviehash, eo->moviesize))
    {
      envy_info_set("Downloading...", 0); 
      if (envy_osdb_download(eo, eo->IDSubtitleFile))
        envy_subtitles_load(eo->subfilename);
    }
  envy_info_set("Logout in OSDB...", 0); 
  envy_osdb_logout(eo);
  envy_info_hide(); 
  envy_osdb_free(eo);

  return;
}

static void
cb_window_resized(Ecore_Evas *ee)
{
  Evas_Coord w, h;

  evas_output_viewport_get(envy->evas, NULL, NULL, &w, &h);
  evas_object_resize(envy->gui, w, h);
  evas_object_resize(envy->black_bg, w, h);
  if (!envy->fullscreen)
    {
      evas_object_move(envy->video, 0, 0);
      evas_object_resize(envy->video, w, h);
    }

  /* Set the correct size for the subtites */

  Evas_Textblock_Style *style;
  char st[512];
  snprintf(st, sizeof(st),
           "DEFAULT='font=vera font_size=%d align=center valign=bottom "
           "color=#ffffff wrap=word style=glow glow_color=#000000'"
           "br='\n'", h/20);
  evas_textblock_style_free((Evas_Textblock_Style *)evas_object_textblock_style_get(envy->sub));
  style = evas_textblock_style_new();
  evas_textblock_style_set(style, st);
  evas_object_textblock_style_set(envy->sub, style);
}

static int
use_video_size(Evas_Object *o)
{
  if (!emotion_object_video_handled_get(o))
    return 1;

  /* Set video size */
  int w, h;
  emotion_object_size_get(o, &w, &h);
  if (w <= 1 && h <= 1) return 1;

  ecore_evas_resize(envy->ee, w, h);
  evas_object_resize(envy->gui, w, h);

  evas_object_hide(envy->balloon);

  return 0;
}

static int
bg_key_down(void *data, int ev_type, void *event_info)
{
  Evas_Object *o = data;
  Ecore_X_Event_Key_Down *ev;

  ev = event_info;
  /* Fullscreen */
  if (!strcmp(ev->keyname, "f"))
    {
      if (!ecore_evas_fullscreen_get(envy->ee))
        {
          envy->fullscreen = 1;
          ecore_evas_fullscreen_set(envy->ee, 1);

          double ratio = 1;
          Ecore_X_Screen_Size size;
          size = ecore_x_randr_current_screen_size_get(ecore_x_window_root_first_get());

          evas_object_move(envy->black_bg, 0, 0);
          evas_object_resize(envy->black_bg, size.width, size.height);

          ratio = emotion_object_ratio_get(o);
          evas_object_resize(o, size.width, size.width/ratio);
          evas_object_move(o, 0, (size.height - size.width/ratio)/2);
          ecore_evas_ignore_events_set(envy->ee, 0);
        }
      else
        {
          envy->fullscreen = 0;
          ecore_evas_fullscreen_set(envy->ee, 0);
        }
    }
  /* Toggle play/pause */
  else if (!strcmp(ev->keyname, "space"))
    {
      if (emotion_object_play_get(data))
        {
          emotion_object_play_set(data, 0);
          edje_object_signal_emit(envy->gui, "pause", "envy");
        }
      else
        {
          emotion_object_play_set(data, 1);
          edje_object_signal_emit(envy->gui, "play", "envy");
        }
    }
  else if (!strcmp(ev->keyname, "Left"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos-10);
    }
  else if (!strcmp(ev->keyname, "Right"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos+10);
    }
  else if (!strcmp(ev->keyname, "Down"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos-60);
    }
  else if (!strcmp(ev->keyname, "Up"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos+60);
    }
  else if (!strcmp(ev->keyname, "Next"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos-600);
    }
  else if (!strcmp(ev->keyname, "Prior"))
    {
      double pos;

      pos = emotion_object_position_get(o);
      emotion_object_position_set(o, pos+600);
    }
  else if (!strcmp(ev->keyname, "q") || !strcmp(ev->keyname, "Escape"))
    {
      exit(EXIT_SUCCESS);
    }
  else if (!strcmp(ev->keyname, "x"))
    {
      char info[512];
      envy->sub_delay += 100./1000.;
      snprintf(info, sizeof(info),  "Sub: %.0f ms", envy->sub_delay*1000);
      envy_info_set(info, 1.0);
    }
  else if (!strcmp(ev->keyname, "y"))
    {
      char info[512];
      envy->sub_delay -= 100./1000.;
      snprintf(info, sizeof(info),  "Sub: %.0f ms", envy->sub_delay*1000);
      envy_info_set(info, 1.0);
    }
  else if (!strcmp(ev->keyname, "s"))
    {
      char info[512];
      if (!envy->subs) return 0;
      envy->sub_delay = envy->subs[0].start - emotion_object_position_get(o);
      snprintf(info, sizeof(info),  "Sub: %.0f ms", envy->sub_delay*1000);
      envy_info_set(info, 1.0);
    }
  else if (!strcmp(ev->keyname, "v"))
    {
      if (envy->show_sub)
        {
          envy->show_sub = 0;
          envy_info_set("Subtitles disabled", 1.0);
        }
      else
        {
          envy->show_sub = 1;
          envy_info_set("Subtitles enabled", 1.0);
        }
    }
  else if (!strcmp(ev->keyname, "d"))
    {
      envy_osdb_auto_search(envy->config->langage, 
                            envy->config->keep_osdb_subfilename);
    }
  else
    {
      _dbg_printf("No bindings for %s\n", ev->keyname);
    }

  return 0;
}

static void
cb_toggle_play_pause(void *data, Evas_Object *o, const char *emission, const char *source)
{
  if (emotion_object_play_get(data))
    {
      emotion_object_play_set(data, 0);
      edje_object_signal_emit(o, "pause", "envy");
    }
  else
    {
      emotion_object_play_set(data, 1);
      edje_object_signal_emit(o, "play", "envy");
    }
}

static void
cb_play(void *data, Evas_Object *o, const char *emission, const char *source)
{
  emotion_object_play_set(data, 1);
  edje_object_signal_emit(o, "play", "envy");
}

static void
cb_pause(void *data, Evas_Object *o, const char *emission, const char *source)
{
  emotion_object_play_set(data, 0);
  edje_object_signal_emit(o, "pause", "envy");
}

#define THE_GOOD_ONE(x, y) (x == 1 ? y : x)
static void
cb_progress_dragged(void *data, Evas_Object *o, const char *emission, const char *source)
{
  if (emotion_object_seekable_get(data))
    {
      double x, y, length;

      edje_object_part_drag_value_get(o, source, &x, &y);
      length = emotion_object_play_length_get(data);
      emotion_object_position_set(data, length*THE_GOOD_ONE(x, y));
    }
  else
    {
      _dbg_printf("Video isn't seekable\n");
    }
}

static void
cb_frame_decoded(void *data, Evas_Object *obj, void *event_info)
{
  double pos, len;
  Sub_Srt *current_sub;

  pos = emotion_object_position_get(obj);
  len = emotion_object_play_length_get(obj);

  /* Set the progress bar at the right place */
  edje_object_part_drag_value_set(envy->gui, "progress", pos/len, pos/len);

  /* Set the timer */
  char str_timer[128];
  sprintf(str_timer, "%02d:%02d:%02d / %02d:%02d:%02d", 
          (int)(pos/3600), (int)(fmod(pos, 3600)/60), (int)(fmod(fmod(pos, 3600), 60)), 
          (int)(len/3600), (int)(fmod(len, 3600)/60), (int)(fmod(fmod(len, 3600), 60)));
  edje_object_part_text_set(envy->gui, "timer", str_timer);

  evas_object_textblock_clear(envy->sub);
  if (!envy->subs || !envy->show_sub) return;

  /* Search for the right subtitle if needed */
  current_sub = envy_subtitle_get(envy->subs, envy->subs_size, pos + envy->sub_delay);
  if (current_sub)
    {
      int w, h;
      int vw, vh;
      evas_object_textblock_text_markup_set(envy->sub, current_sub->txt);
      evas_object_textblock_size_formatted_get(envy->sub, &w, &h);
      ecore_evas_geometry_get(envy->ee, NULL, NULL, &vw, &vh);
      evas_object_resize(envy->sub, vw, h);
      evas_object_move(envy->sub, 0, vh - h - envy->config->sub_padding);
    } 
}

static void 
cb_mouse_down(void *data, Evas *e, Evas_Object *o, void *event_info)
{
  if (envy->mouse_timer) 
    {
      ecore_timer_del(envy->mouse_timer);
      envy->mouse_timer = NULL;
    }
}

static void 
cb_mouse_move(void *data, Evas *e, Evas_Object *o, void *event_info)
{
  Ecore_Evas *ee = data;
  Ecore_X_Window win;
  win = envy_engine_window_get(envy->config->ecore_evas_engine_type, ee);
  ecore_x_window_cursor_show(win, 1);

  if (envy->ui_visible == 0) 
    {
      edje_object_signal_emit(envy->gui, "show_gui", "envy");
      envy->ui_visible = 1;
    }

  if (envy->mouse_timer) ecore_timer_del(envy->mouse_timer);
  envy->mouse_timer = ecore_timer_add(2.0, cb_hide_mouse, ee);
}

static int
cb_hide_info(void *data)
{
  envy_info_hide();
  envy->info_timer = NULL;
  return 0;
}

static int
cb_hide_mouse(void *data)
{
  Ecore_X_Window win;
  win = envy_engine_window_get(envy->config->ecore_evas_engine_type, data);
  ecore_x_window_cursor_show(win, 0);
  envy->mouse_timer = NULL;
  edje_object_signal_emit(envy->gui, "hide_gui", "envy");
  envy->ui_visible = 0;
  return 0;
}

static void
envy_info_set(const char *msg, double time)
{
  if (envy->info_timer) ecore_timer_del(envy->info_timer);
  envy->info_timer = NULL;
  if (time != 0) envy->info_timer = ecore_timer_add(time, cb_hide_info, NULL);
  edje_object_part_text_set(envy->gui, "info", msg);
  edje_object_signal_emit(envy->gui, "show_info", "envy");
}

static void
envy_info_hide(void)
{
  edje_object_part_text_set(envy->gui, "info", NULL);
  edje_object_signal_emit(envy->gui, "hide_info", "envy");
}
